import { idText, isLiteralExpression, isPropertyAccessExpression, isIdentifier, getCombinedModifierFlags, getJSDocTags, SyntaxKind, isPropertyDeclaration, isPropertySignature, isIndexSignatureDeclaration, isMethodDeclaration, isMethodSignature, isConstructorDeclaration, isConstructSignatureDeclaration, isCallSignatureDeclaration, isTypeReferenceNode, isTypeLiteralNode, isArrayTypeNode, isTupleTypeNode, isParenthesizedTypeNode, isTypeOperatorNode, isTypePredicateNode, isUnionTypeNode, isIntersectionTypeNode, isMappedTypeNode, isConditionalTypeNode, isIndexedAccessTypeNode, isFunctionOrConstructorTypeNode, isImportTypeNode, isTypeQueryNode, isInferTypeNode, isLiteralTypeNode, isStringLiteral, isModuleBlock, isEnumDeclaration, isFunctionDeclaration, isClassDeclaration, isInterfaceDeclaration, isTypeAliasDeclaration, isModuleDeclaration, isVariableStatement, isExportDeclaration, ModifierFlags, isExportAssignment, isImportDeclaration, isNamedImports, ModuleKind, nodeModuleNameResolver, sys, createSourceFile, ScriptTarget, createProgram as createProgram$1, findConfigFile, getParsedCommandLineOfConfigFile } from 'typescript';
import path from 'path';

const MARKER = "0";
let IDs = 1;
/**
 * Create a new `Program` for the given `node`:
 */
function createProgram(node) {
    return withStartEnd({
        type: "Program",
        sourceType: "module",
        body: [],
    }, node);
}
/**
 * Create a default export for `id`:
 * `export default id`
 */
function createDefaultExport(node, range) {
    return withStartEnd({
        type: "ExportDefaultDeclaration",
        declaration: createIdentifier(node),
    }, range);
}
/**
 * Create an export for `id`:
 * `export { id }`
 */
function createExport(node, range) {
    const id = createIdentifier(node);
    return withStartEnd({
        type: "ExportNamedDeclaration",
        declaration: null,
        specifiers: [
            {
                type: "ExportSpecifier",
                exported: id,
                local: id,
            },
        ],
    }, range);
}
/**
 * Creates a reference to `id`:
 * `_ = ${id}`
 */
function createReference(id) {
    return {
        type: "AssignmentPattern",
        left: {
            type: "Identifier",
            name: String(IDs++),
        },
        right: id,
    };
}
function createIdentifier(node) {
    return withStartEnd({
        type: "Identifier",
        name: node.getText(),
    }, {
        start: node.getStart(),
        end: node.getEnd(),
    });
}
/**
 * Create a new Declaration and Scope for `id`:
 * `function ${id}(_ = MARKER) {}`
 */
function createDeclaration(id, range) {
    return withStartEnd({
        type: "FunctionDeclaration",
        id: withStartEnd({
            type: "Identifier",
            name: idText(id),
        }, { start: id.getStart(), end: id.getEnd() }),
        params: [],
        body: { type: "BlockStatement", body: [] },
    }, range);
}
function convertExpression(node) {
    if (isLiteralExpression(node)) {
        return { type: "Literal", value: node.text };
    }
    if (isPropertyAccessExpression(node)) {
        return withStartEnd({
            type: "MemberExpression",
            computed: false,
            object: convertExpression(node.expression),
            property: convertExpression(node.name),
        }, {
            start: node.expression.getStart(),
            end: node.name.getEnd(),
        });
    }
    // istanbul ignore else
    if (isIdentifier(node)) {
        return createIdentifier(node);
    }
    else {
        console.log({ kind: node.kind, code: node.getFullText() });
        throw new Error(`Unknown Expression`);
    }
}
/**
 * Mark the nested `range` to be removed, by creating dead code:
 * `_ = () => {MARKER}`
 */
function removeNested(range) {
    return createReference({
        type: "FunctionExpression",
        id: null,
        body: withStartEnd({
            type: "BlockStatement",
            body: [
                withStartEnd({
                    type: "ExpressionStatement",
                    expression: { type: "Identifier", name: MARKER },
                }, range),
            ],
        }, {
            start: getStart(range) - 1,
            end: range.end + 1,
        }),
        params: [],
    });
}
function getStart({ start, pos }) {
    return typeof start === "number" ? start : pos || 0;
}
function withStartEnd(node, range) {
    return Object.assign(node, {
        start: getStart(range),
        end: range.end,
    });
}
function matchesModifier(node, flags) {
    const nodeFlags = getCombinedModifierFlags(node);
    return (nodeFlags & flags) === flags;
}
function isInternal(node) {
    const tags = getJSDocTags(node);
    return tags.some(t => idText(t.tagName) === "internal");
}

const IGNORE_TYPENODES = new Set([
    SyntaxKind.LiteralType,
    SyntaxKind.VoidKeyword,
    SyntaxKind.UnknownKeyword,
    SyntaxKind.AnyKeyword,
    SyntaxKind.BooleanKeyword,
    SyntaxKind.NumberKeyword,
    SyntaxKind.StringKeyword,
    SyntaxKind.ObjectKeyword,
    SyntaxKind.NullKeyword,
    SyntaxKind.UndefinedKeyword,
    SyntaxKind.SymbolKeyword,
    SyntaxKind.NeverKeyword,
    SyntaxKind.ThisKeyword,
    SyntaxKind.ThisType,
]);
let inlineImports = 1;
class DeclarationScope {
    constructor({ id, range, transformer }) {
        /**
         * As we walk the AST, we need to keep track of type variable bindings that
         * shadow the outer identifiers. To achieve this, we keep a stack of scopes,
         * represented as Sets of variable names.
         */
        this.scopes = [];
        this.transformer = transformer;
        this.declaration = createDeclaration(id, range);
    }
    pushScope() {
        this.scopes.push(new Set());
    }
    popScope(n = 1) {
        for (let i = 0; i < n; i++) {
            this.scopes.pop();
        }
    }
    pushTypeVariable(id) {
        const name = id.getText();
        this.scopes[this.scopes.length - 1].add(name);
    }
    pushRaw(expr) {
        this.declaration.params.push(expr);
    }
    pushReference(id) {
        let name;
        // We convert references from TS AST to ESTree
        // to hand them off to rollup.
        // This means we have to check the left-most identifier inside our scope
        // tree and avoid to create the reference in that case
        if (id.type === "Identifier") {
            name = id.name;
        }
        else if (id.type === "MemberExpression") {
            if (id.object.type === "Identifier") {
                name = id.object.name;
            }
        }
        if (name) {
            for (const scope of this.scopes) {
                if (scope.has(name)) {
                    return;
                }
            }
        }
        this.pushRaw(createReference(id));
    }
    pushIdentifierReference(id) {
        this.pushReference(createIdentifier(id));
    }
    removeModifier(node, kind = SyntaxKind.ExportKeyword) {
        for (const mod of node.modifiers || []) {
            if (mod.kind === kind) {
                const start = node.getStart();
                const end = mod.end + 1;
                this.pushRaw(removeNested({ start, end }));
            }
        }
    }
    convertParametersAndType(node) {
        const typeVariables = this.convertTypeParameters(node.typeParameters);
        for (const param of node.parameters) {
            this.convertTypeNode(param.type);
        }
        this.convertTypeNode(node.type);
        this.popScope(typeVariables);
    }
    convertHeritageClauses(node) {
        for (const heritage of node.heritageClauses || []) {
            for (const type of heritage.types) {
                this.pushReference(convertExpression(type.expression));
                for (const arg of type.typeArguments || []) {
                    this.convertTypeNode(arg);
                }
            }
        }
    }
    convertMembers(members) {
        for (const node of members) {
            // NOTE(swatinem):
            // Well, actually having `private`/`protected` members in the exported
            // definitions is quite nice, so let’s keep them. Instead we look for an
            // `@internal` tag
            // if (matchesModifier(node, ts.ModifierFlags.Private) || matchesModifier(node, ts.ModifierFlags.Protected)) {
            if (isInternal(node)) {
                this.pushRaw(removeNested({ start: node.getFullStart(), end: node.getEnd() }));
                continue;
            }
            if (isPropertyDeclaration(node) || isPropertySignature(node) || isIndexSignatureDeclaration(node)) {
                this.convertTypeNode(node.type);
                continue;
            }
            // istanbul ignore else
            if (isMethodDeclaration(node) ||
                isMethodSignature(node) ||
                isConstructorDeclaration(node) ||
                isConstructSignatureDeclaration(node) ||
                isCallSignatureDeclaration(node)) {
                this.convertParametersAndType(node);
            }
            else {
                console.log({ kind: node.kind, code: node.getFullText() });
                throw new Error(`Unknown TypeElement`);
            }
        }
    }
    convertTypeParameters(params) {
        if (!params) {
            return 0;
        }
        for (const node of params) {
            this.pushScope();
            this.pushTypeVariable(node.name);
            this.convertTypeNode(node.default);
        }
        return params.length;
    }
    convertTypeNode(node) {
        if (!node) {
            return;
        }
        if (IGNORE_TYPENODES.has(node.kind)) {
            return;
        }
        if (isTypeReferenceNode(node)) {
            this.pushReference(this.convertEntityName(node.typeName));
            for (const arg of node.typeArguments || []) {
                this.convertTypeNode(arg);
            }
            return;
        }
        if (isTypeLiteralNode(node)) {
            return this.convertMembers(node.members);
        }
        if (isArrayTypeNode(node)) {
            return this.convertTypeNode(node.elementType);
        }
        if (isTupleTypeNode(node)) {
            for (const type of node.elementTypes) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isParenthesizedTypeNode(node) || isTypeOperatorNode(node) || isTypePredicateNode(node)) {
            return this.convertTypeNode(node.type);
        }
        if (isUnionTypeNode(node) || isIntersectionTypeNode(node)) {
            for (const type of node.types) {
                this.convertTypeNode(type);
            }
            return;
        }
        if (isMappedTypeNode(node)) {
            const { typeParameter, type } = node;
            this.convertTypeNode(typeParameter.constraint);
            this.pushScope();
            this.pushTypeVariable(node.typeParameter.name);
            this.convertTypeNode(type);
            this.popScope();
            return;
        }
        if (isConditionalTypeNode(node)) {
            this.convertTypeNode(node.checkType);
            this.pushScope();
            this.convertTypeNode(node.extendsType);
            this.convertTypeNode(node.trueType);
            this.convertTypeNode(node.falseType);
            this.popScope();
            return;
        }
        if (isIndexedAccessTypeNode(node)) {
            this.convertTypeNode(node.objectType);
            this.convertTypeNode(node.indexType);
            return;
        }
        if (isFunctionOrConstructorTypeNode(node)) {
            this.convertParametersAndType(node);
            return;
        }
        if (isImportTypeNode(node)) {
            this.convertImportTypeNode(node);
            return;
        }
        if (isTypeQueryNode(node)) {
            this.pushReference(this.convertEntityName(node.exprName));
            return;
        }
        // istanbul ignore else
        if (isInferTypeNode(node)) {
            this.pushTypeVariable(node.typeParameter.name);
            return;
        }
        else {
            console.log(node.getSourceFile().getFullText());
            console.log({ kind: node.kind, code: node.getFullText() });
            throw new Error(`Unknown TypeNode`);
        }
    }
    // For import type nodes of the form
    // `import("./foo").Bar`
    // we create the following ESTree equivalent:
    // 1. `import * as _ from "./foo";` on the toplevel
    // 2. `_.Bar` in our declaration scope
    convertImportTypeNode(node) {
        // istanbul ignore if
        if (!isLiteralTypeNode(node.argument) || !isStringLiteral(node.argument.literal)) {
            console.log({ kind: node.kind, code: node.getFullText() });
            throw new Error("inline imports should have a literal argument");
        }
        const fileId = node.argument.literal.text;
        const importId = String(inlineImports++);
        const importIdRef = withStartEnd({
            type: "Identifier",
            name: importId,
        }, {
            start: node.getStart(),
            end: node.getEnd(),
        });
        this.transformer.pushStatement({
            type: "ImportDeclaration",
            specifiers: [
                {
                    type: "ImportNamespaceSpecifier",
                    local: { type: "Identifier", name: importId },
                },
            ],
            source: { type: "Literal", value: fileId },
        });
        if (node.qualifier && isIdentifier(node.qualifier)) {
            this.pushReference(withStartEnd({
                type: "MemberExpression",
                computed: false,
                object: importIdRef,
                property: createIdentifier(node.qualifier),
            }, { start: node.getStart(), end: node.getEnd() }));
        }
        else {
            // we definitely need to do some string manipulation on the source code,
            // since rollup will not touch the `import("...")` bit at all.
            // also, for *internal* namespace references, we have the same problem
            // as with re-exporting references… -_-
            this.pushReference(importIdRef);
        }
    }
    convertEntityName(node) {
        if (isIdentifier(node)) {
            return createIdentifier(node);
        }
        return withStartEnd({
            type: "MemberExpression",
            computed: false,
            object: this.convertEntityName(node.left),
            property: createIdentifier(node.right),
        }, 
        // TODO: clean up all the `start` handling!
        { start: node.getStart(), end: node.end });
    }
    convertNamespace(node) {
        this.pushScope();
        // istanbul ignore if
        if (!node.body || !isModuleBlock(node.body)) {
            console.log({ code: node.getFullText() });
            throw new Error(`namespace must have a "ModuleBlock" body.`);
        }
        const { statements } = node.body;
        // first, hoist all the declarations for correct shadowing
        for (const stmt of statements) {
            if (isEnumDeclaration(stmt) ||
                isFunctionDeclaration(stmt) ||
                isClassDeclaration(stmt) ||
                isInterfaceDeclaration(stmt) ||
                isTypeAliasDeclaration(stmt) ||
                isModuleDeclaration(stmt)) {
                // istanbul ignore else
                if (stmt.name && isIdentifier(stmt.name)) {
                    this.pushTypeVariable(stmt.name);
                }
                else {
                    console.log({ code: stmt.getFullText() });
                    throw new Error(`non-Identifier name not supported`);
                }
                continue;
            }
            if (isVariableStatement(stmt)) {
                for (const decl of stmt.declarationList.declarations) {
                    // istanbul ignore else
                    if (isIdentifier(decl.name)) {
                        this.pushTypeVariable(decl.name);
                    }
                    else {
                        console.log({ code: decl.getFullText() });
                        throw new Error(`non-Identifier name not supported`);
                    }
                }
                continue;
            }
            // istanbul ignore else
            if (isExportDeclaration(stmt)) ;
            else {
                console.log({ code: stmt.getFullText() });
                throw new Error(`namespace child (hoisting) not supported yet`);
            }
        }
        // and then walk all the children like normal…
        for (const stmt of statements) {
            if (isVariableStatement(stmt)) {
                for (const decl of stmt.declarationList.declarations) {
                    if (decl.type) {
                        this.convertTypeNode(decl.type);
                    }
                }
                continue;
            }
            if (isFunctionDeclaration(stmt)) {
                this.convertParametersAndType(stmt);
                continue;
            }
            if (isInterfaceDeclaration(stmt) || isClassDeclaration(stmt)) {
                const typeVariables = this.convertTypeParameters(stmt.typeParameters);
                this.convertHeritageClauses(stmt);
                this.convertMembers(stmt.members);
                this.popScope(typeVariables);
                continue;
            }
            if (isTypeAliasDeclaration(stmt)) {
                const typeVariables = this.convertTypeParameters(stmt.typeParameters);
                this.convertTypeNode(stmt.type);
                this.popScope(typeVariables);
                continue;
            }
            if (isModuleDeclaration(stmt)) {
                this.convertNamespace(stmt);
                continue;
            }
            if (isEnumDeclaration(stmt)) {
                // noop
                continue;
            }
            // istanbul ignore else
            if (isExportDeclaration(stmt)) {
                if (stmt.exportClause) {
                    for (const decl of stmt.exportClause.elements) {
                        const id = decl.propertyName || decl.name;
                        this.pushIdentifierReference(id);
                    }
                }
            }
            else {
                console.log({ code: stmt.getFullText() });
                throw new Error(`namespace child (walking) not supported yet`);
            }
        }
        this.popScope();
    }
}

class Transformer {
    constructor(sourceFile) {
        this.exports = new Set();
        this.ast = createProgram(sourceFile);
        for (const stmt of sourceFile.statements) {
            this.convertStatement(stmt);
        }
    }
    transform() {
        return this.ast;
    }
    pushStatement(node) {
        this.ast.body.push(node);
    }
    maybeMarkAsExported(node, id) {
        if (isInternal(node)) {
            return false;
        }
        if (matchesModifier(node, ModifierFlags.ExportDefault)) {
            const start = node.pos;
            this.pushStatement(createDefaultExport(id, { start, end: start }));
            return true;
        }
        else if (matchesModifier(node, ModifierFlags.Export)) {
            const start = node.pos;
            const name = id.getText();
            if (this.exports.has(name)) {
                return true;
            }
            this.pushStatement(createExport(id, { start, end: start }));
            this.exports.add(name);
            return true;
        }
        return false;
    }
    createDeclaration(id, range) {
        const scope = new DeclarationScope({ id, range, transformer: this });
        this.pushStatement(scope.declaration);
        return scope;
    }
    convertStatement(node) {
        if (isEnumDeclaration(node)) {
            return this.convertEnumDeclaration(node);
        }
        if (isFunctionDeclaration(node)) {
            return this.convertFunctionDeclaration(node);
        }
        if (isInterfaceDeclaration(node) || isClassDeclaration(node)) {
            return this.convertClassOrInterfaceDeclaration(node);
        }
        if (isTypeAliasDeclaration(node)) {
            return this.convertTypeAliasDeclaration(node);
        }
        if (isVariableStatement(node)) {
            return this.convertVariableStatement(node);
        }
        if (isExportDeclaration(node) || isExportAssignment(node)) {
            return this.convertExportDeclaration(node);
        }
        if (isModuleDeclaration(node)) {
            return this.convertNamespaceDeclaration(node);
        }
        // istanbul ignore else
        if (isImportDeclaration(node)) {
            return this.convertImportDeclaration(node);
        }
        else {
            console.log({ kind: node.kind, code: node.getFullText() });
            throw new Error(`unsupported node type`);
        }
    }
    convertNamespaceDeclaration(node) {
        // istanbul ignore if
        if (!isIdentifier(node.name)) {
            console.log({ code: node.getFullText() });
            throw new Error(`namespace name should be an "Identifier"`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
        scope.convertNamespace(node);
    }
    convertEnumDeclaration(node) {
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
    }
    convertFunctionDeclaration(node) {
        // istanbul ignore if
        if (!node.name) {
            console.log({ code: node.getFullText() });
            throw new Error(`FunctionDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        scope.pushIdentifierReference(node.name);
        scope.convertParametersAndType(node);
    }
    convertClassOrInterfaceDeclaration(node) {
        // istanbul ignore if
        if (!node.name) {
            console.log({ code: node.getFullText() });
            throw new Error(`ClassDeclaration / InterfaceDeclaration should have a name`);
        }
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        if (isInterfaceDeclaration(node)) {
            scope.removeModifier(node, SyntaxKind.DefaultKeyword);
        }
        const typeVariables = scope.convertTypeParameters(node.typeParameters);
        scope.convertHeritageClauses(node);
        scope.convertMembers(node.members);
        scope.popScope(typeVariables);
    }
    convertTypeAliasDeclaration(node) {
        this.maybeMarkAsExported(node, node.name);
        const scope = this.createDeclaration(node.name, node);
        scope.removeModifier(node);
        const typeVariables = scope.convertTypeParameters(node.typeParameters);
        scope.convertTypeNode(node.type);
        scope.popScope(typeVariables);
    }
    convertVariableStatement(node) {
        const { declarations } = node.declarationList;
        // istanbul ignore if
        if (declarations.length !== 1) {
            console.log({ code: node.getFullText() });
            throw new Error(`VariableStatement with more than one declaration not yet supported`);
        }
        for (const decl of declarations) {
            // istanbul ignore if
            if (!isIdentifier(decl.name)) {
                console.log({ code: node.getFullText() });
                throw new Error(`VariableDeclaration must have a name`);
            }
            this.maybeMarkAsExported(node, decl.name);
            const scope = this.createDeclaration(decl.name, node);
            scope.removeModifier(node);
            scope.convertTypeNode(decl.type);
        }
    }
    convertExportDeclaration(node) {
        if (isExportAssignment(node)) {
            this.pushStatement(withStartEnd({
                type: "ExportDefaultDeclaration",
                declaration: convertExpression(node.expression),
            }, node));
            return;
        }
        const source = node.moduleSpecifier ? convertExpression(node.moduleSpecifier) : undefined;
        if (!node.exportClause) {
            this.pushStatement(withStartEnd({
                type: "ExportAllDeclaration",
                source,
            }, node));
        }
        else {
            this.pushStatement(withStartEnd({
                type: "ExportNamedDeclaration",
                declaration: null,
                specifiers: node.exportClause
                    ? node.exportClause.elements.map(e => this.convertExportSpecifier(e))
                    : /* istanbul ignore next */ [],
                source,
            }, node));
        }
    }
    convertImportDeclaration(node) {
        const source = convertExpression(node.moduleSpecifier);
        if (!node.importClause) {
            return;
        }
        // istanbul ignore if
        if (!node.importClause.name && !node.importClause.namedBindings) {
            console.log({ code: node.getFullText() });
            throw new Error(`ImportDeclaration should have imports`);
        }
        const specifiers = node.importClause.namedBindings
            ? this.convertNamedImportBindings(node.importClause.namedBindings)
            : [];
        if (node.importClause.name) {
            specifiers.push({
                type: "ImportDefaultSpecifier",
                local: createIdentifier(node.importClause.name),
            });
        }
        this.pushStatement(withStartEnd({
            type: "ImportDeclaration",
            specifiers,
            source,
        }, node));
    }
    convertNamedImportBindings(node) {
        if (isNamedImports(node)) {
            return node.elements.map(el => {
                const local = createIdentifier(el.name);
                const imported = el.propertyName ? createIdentifier(el.propertyName) : local;
                return {
                    type: "ImportSpecifier",
                    local,
                    imported,
                };
            });
        }
        return [
            {
                type: "ImportNamespaceSpecifier",
                local: createIdentifier(node.name),
            },
        ];
    }
    convertExportSpecifier(node) {
        const exported = createIdentifier(node.name);
        return {
            type: "ExportSpecifier",
            exported: exported,
            local: node.propertyName ? createIdentifier(node.propertyName) : exported,
        };
    }
}

let SOURCEMAPPING_URL = "sourceMa";
SOURCEMAPPING_URL += "ppingURL";
const SOURCEMAPPING_URL_RE = new RegExp(`^//#\\s+${SOURCEMAPPING_URL}=.+\\n?`, "m");
var CompileMode;
(function (CompileMode) {
    CompileMode["Types"] = "dts";
    CompileMode["Js"] = "js";
})(CompileMode || (CompileMode = {}));
const OPTIONS_OVERRIDES = {
    module: ModuleKind.ES2015,
    noEmitOnError: false,
    noEmit: false,
    skipLibCheck: true,
    declaration: true,
    allowJs: true,
    checkJs: true,
    resolveJsonModule: true,
    sourceMap: true,
    inlineSourceMap: false,
    declarationMap: true,
};
const COMPILERCACHE = new Map();
function createCompiler(options) {
    const file = path.basename(options.tsconfig);
    const configFileName = findConfigFile(options.tsconfig, sys.fileExists, path.extname(options.tsconfig) ? file : undefined);
    // istanbul ignore if
    if (!configFileName) {
        throw new Error(`rollup-plugin-dts: Couldn't find tsconfig file`);
    }
    const compilerOptions = Object.assign({}, options.compilerOptions, OPTIONS_OVERRIDES);
    let diagnostic;
    const configParseResult = getParsedCommandLineOfConfigFile(configFileName, compilerOptions, Object.assign({}, sys, { onUnRecoverableConfigFileDiagnostic(d) {
            // istanbul ignore next
            diagnostic = d;
        } }));
    // istanbul ignore if
    if (!configParseResult) {
        console.log(diagnostic);
        throw new Error(`rollup-plugin-dts: Couldn't process compiler options`);
    }
    const { fileNames, options: parsedCompilerOptions } = configParseResult;
    return {
        parsedCompilerOptions,
        compiler: createProgram$1(fileNames, parsedCompilerOptions),
    };
}
function getEmitFiles(compiler, fileName) {
    const result = {
        dts: "",
        dtsMap: `{"mappings": ""}`,
        js: "",
        jsMap: `{"mappings": ""}`,
    };
    if (fileName.endsWith(".d.ts")) {
        result.dts = sys.readFile(fileName, "utf-8");
        return result;
    }
    const sourceFile = compiler.getSourceFile(fileName);
    // XXX(swatinem): maybe we should look at the diagnostics? :-D
    compiler.emit(sourceFile, (fileName, data) => {
        data = data.replace(SOURCEMAPPING_URL_RE, "").trim();
        if (fileName.endsWith(".d.ts")) {
            result.dts = data;
        }
        else if (fileName.endsWith(".js")) {
            result.js = data;
            // NOTE(swatinem): hm, there are still issues with this,
            // I couldn’t get it to work locally
            // See https://github.com/Microsoft/TypeScript/issues/25662
            // } else if (fileName.endsWith(".d.ts.map")) {
            //   result.dtsMap = data;
        }
        else if (fileName.endsWith(".js.map")) {
            result.jsMap = data;
        }
    });
    return result;
}
function getCachedCompiler(options) {
    const cacheKey = JSON.stringify(options);
    function lazyCreate() {
        let compiler = COMPILERCACHE.get(cacheKey);
        if (!compiler) {
            compiler = createCompiler(options);
            COMPILERCACHE.set(cacheKey, compiler);
        }
        return compiler;
    }
    return {
        resolve(importee, importer) {
            const { parsedCompilerOptions } = lazyCreate();
            const result = nodeModuleNameResolver(importee, importer, parsedCompilerOptions, sys);
            // hm, maybe we should use `isExternalLibraryImport` at some point…
            // istanbul ignore else
            if (result.resolvedModule && result.resolvedModule.resolvedFileName) {
                return result.resolvedModule.resolvedFileName;
            }
            else {
                return;
            }
        },
        load(fileName) {
            const { compiler } = lazyCreate();
            const emitFiles = getEmitFiles(compiler, fileName);
            if (options.mode === CompileMode.Js) {
                return {
                    code: emitFiles.js,
                    map: emitFiles.jsMap,
                };
            }
            let code = emitFiles.dts;
            let dtsFileName = fileName;
            if (!fileName.endsWith(".d.ts")) {
                dtsFileName = fileName.slice(0, -path.extname(fileName).length) + ".d.ts";
            }
            const dtsSource = createSourceFile(dtsFileName, code, ScriptTarget.Latest, true);
            const converter = new Transformer(dtsSource);
            const ast = converter.transform();
            // NOTE(swatinem):
            // hm, typescript generates `export default` without a declare,
            // but rollup moves the `export default` to a different place, which leaves
            // the function declaration without a `declare`.
            // Well luckily both words have the same length, haha :-D
            code = code.replace(/(export\s+)default(\s+(function|class))/m, "$1declare$2");
            return {
                code,
                ast,
                map: emitFiles.dtsMap,
            };
        },
    };
}

const version = "0.14.0";

const TSLIB_ID = "\0tslib";
const BANNER = `
// FILE GENERATED BY \`rollup-plugin-dts@${version}\`
// https://github.com/Swatinem/rollup-plugin-dts
`.trim() + "\n";
const plugin = (options = {}) => {
    const tslibFileName = require.resolve("tslib").replace("tslib.js", "tslib.es6.js");
    const tslib = sys.readFile(tslibFileName, "utf-8") || null;
    const mode = options.compileMode || CompileMode.Types;
    const compiler = getCachedCompiler({
        tsconfig: options.tsconfig || process.cwd(),
        compilerOptions: options.compilerOptions || {},
        mode,
    });
    return {
        name: "dts",
        banner: mode === CompileMode.Types && options.banner !== false ? BANNER : undefined,
        outputOptions(options) {
            if (mode === CompileMode.Js) {
                return options;
            }
            return Object.assign({}, options, { format: "es", exports: "named", compact: false, freeze: false, interop: false, namespaceToStringTag: false, strict: false });
        },
        resolveId(importee, importer) {
            // istanbul ignore if
            if (importee === "tslib") {
                return TSLIB_ID;
            }
            if (!importer) {
                return;
            }
            importer = importer.split("\\").join("/");
            return compiler.resolve(importee, importer);
        },
        async load(id) {
            if (id === TSLIB_ID) {
                return tslib;
            }
            if (id.endsWith(".ts") || id.endsWith(".tsx")) {
                return compiler.load(id);
            }
            return null;
        },
    };
};
const dts = (options = {}) => {
    options.compileMode = options.compileMode || CompileMode.Types;
    return plugin(options);
};
const js = (options = {}) => {
    options.compileMode = options.compileMode || CompileMode.Js;
    return plugin(options);
};

export default plugin;
export { CompileMode, dts, js, plugin, js as ts };
